/*

  Text-based MP3 / Ogg Vorbis Encoder demonstration for VS1205 / VS1005.
  VLSI Solution 2015-07-17 HH

  THIS IS VERY MUCH WORK-IN-PROGRESS, SO DON'T READ TOO MUCH INTO THE
  CODE AND COMMENTS BELOW!!!!! - 2015-07-17 Henrik


  This VSOS application reads audio from VS1005g's default input
  (typically line input) and writes it as an MP3/OGG file to a given file.

  - To use microphone inputs, set USE_MIC_INPUTS below.
  - To set automatic gain control, set USE_AGC below.
    Note that the AGC library is needed even if AGC itself is not used.
  - To unset Subsonic filter if AGC is not used, unset USE_SUBSONIC below.
  - To amplify the signal with a constant, set AMPLIFY_SIGNAL_BY below.
  

  IMPORTANT!!!

  To properly work, this program requires the FILEBUF.DL3 dynamic
  library to be installed first. If you are using VSOS 3.12 or
  older, you can compile the dynamic library yourself. If using
  VSOS 3.13 or newer, the library .h files are included in the VSOS
  release. Using the library, this program allocates the largest
  contiguous Y memory chunk it can get. If it can get 38 kilobytes
  or more for buffering purposes, 256 kbps or lower MP3 files don't
  suffer from interruptions with SD cards tested by VLSI.

  This program requires the ENCMP3.DL3 dynamic library to be installed
  if you want to encode MP3 files. If you are using VSOS 3.12 or older,
  you can compile the dynamic library yourself.

  This program requires the ENCVORB.DL3 dynamic library to be installed
  if you want to encode Ogg Vorbis files. If you are using VSOS 3.12 or
  older, you can compile the dynamic library yourself.

  Unless you disable both Automatic Gain Control and the Subsonic Filter,
  the AGC.DL3 dynamic library needs to be installed. If you are using
  VSOS 3.12 or older, you can compile the dynamic library yourself.


  DISCLAIMER

  This code may be used freely in any product containing one or more ICs
  by VLSI Solution.

  No guarantee is given for the usability of this code.
  
*/
#include <vo_stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sysmemory.h>
#include <encoder.h>
#include <audiofs.h>
#include <devAudio.h>
#include <ringbuf.h>
#include <vsostasks.h>
#include <timers.h>
#include <stdbuttons.h>
#include <audio.h>
#include <lcd.h>
#include <rgb565.h>
#include <vsos.h>
#include <apploader.h>
#include <libfilebuf.h>
#include <autobuffer.h>
#include <saturate.h>
#include <string.h>
#include <libagc.h>
#include <agc.h>
#include <sysmemory.h>
#include <vo_gpio.h>
#include <uimessages.h>
#include <consolestate.h>
#include "libencmp3.h"
#include "libencvorb.h"


/* ----- USER ADJUSTABLE PARAMETERS START HERE!!! ----- */


enum ChannelMode {
  chmJointStereo = 0,	/* Normal stereo */
  chmLeft = 2,		/* Mono, use only left channel */
  chmRight,		/* Mono, use only right channel */
  chmMono		/* Mono downmix of left & right */
};

/*
  To use Line in / Mic, and if you are using the VS1005 Developer Board,
  need to adjust some jumpers if you also need the right channel:
  Mic:
  - Move JP5 jumper left to MIC2 pins 3 & 2.
  - Insert JP6 jumper left to MIC2 pins 3 & 2.
  Line in:
  - Move JP5 jumper right to LINE2 pins 2 & 1.
  - Remove JP6 jumper.
  Note that AGC is very useful when using the microphone input.
*/

/*
  Make this #if 1 if you want to use Automatic Gain Control (AGC)
  AGC_MAX_GAIN may vary between 1.0 and 64.0.
  Recommended value is somewhere between 2.0 (+6 dB) and 8.0 (+18 dB).
  With higher values microphone background noise may become
  disturbing.
*/
#if 0
#  define USE_AGC
#  define AGC_MAX_GAIN (4.0) /* Value/expression should be in parenthesis! */
#else
   /* If you are not using AGC, you can disable the Subsonic Filter by
      putting #if 0 here. That way the AGC library is not needed. For best
      sound quality using the Subsonic Filter is highly recommended. */
#  if 1
#    define USE_SUBSONIC
#  endif
#endif

/*
  Make this "#if 1" if you want to digitally amplify the input
  signal before encoding it to MP3.
  If using an amplification of more than 1, you run the risk of
  distorting strong signals, but this may be useful if your
  microphone signal is "always too low". The best way to solve
  this issue usually is to use an AGC filter.

  The amount of amplification is linear, so 1.0 = unchanged,
  2.0 = +6 dB, 4.0 = +12 dB, etc. Maximum value is 64, but in
  practise you should stay well below that or you might end up
  hearing noise and distortion.
 */
#if 0
#define AMPLIFY_SIGNAL_BY (2.0) /* Value/expression should be in parenthesis! */
#endif

/* ----- USER ADJUSTABLE PARAMETERS END HERE!!! ----- */






#ifndef MIN
#  define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef MAX
#  define MAX(a,b) (((a)>(b))?(a):(b))
#endif

FILE *oFp = NULL;       ///< Output file pointer
struct Encoder *enc; ///< Encoder structure
enum ChannelMode channelMode = chmJointStereo;

u_int16 vuMeterMemory[2] = {0, 0};
u_int16 clearVuMeter = 1;
u_int16 pauseMode = 0;

#if defined(USE_SUBSONIC) || defined(USE_AGC)
#  define MAX_AGC_CHANNELS 2
#  include <libagc.h>
struct SubsonicM subsonicM[MAX_AGC_CHANNELS];
struct SubsonicG __mem_y subsonicG;
#  ifdef USE_AGC
struct Agc agc[MAX_AGC_CHANNELS];
struct AgcConsts __mem_y agcConsts;
#  endif /* USE_AGC */
void *agcLib = NULL;
#endif /* AGC_SUBSONIC */

auto void VuMeter(register __i2 const s_int16 *d, register __d0 u_int16 n,
                  register __i1 u_int16 *mem) {
  u_int16 memCopy = *mem;
  u_int16 i;
  for (i=0; i<n; i++) {
    u_int16 t = abs(*d++);
    if (t > memCopy) {
      memCopy = t;
    }
  }
  *mem = memCopy;
}

#ifdef AMPLIFY_SIGNAL_BY
auto void Gain(register __i2 s_int16 *d, register __d0 u_int16 n,
	       register __c0 u_int16 gainX1024) {  
  int i;
  for (i=0; i<n; i++) {
      *d = Sat32To16((s_int32)(*d) * gainX1024 >> 10);
      d++;
  }
}
#endif

/**
   struct EncoderServices service ESRead(). Reads bytes into an u_int16
   table.

   \param es Encoder Service structure.
   \param data Pointer to destination data pointers, one for each channel.
   \param samples How many n-channel 16-bit samples to read.
        word is only half filled, only the MSB is changed.
   \param chan Number of channels.
   \return Number of samples read. 0 is an EOF or error.
*/
u_int16 ESRead(struct EncoderServices *es, s_int16 **data, u_int16 samples,
               u_int16 chan) {
  u_int16 i;
#define STB 64
  static s_int16 stb[STB]; /* Temporary buffer */

  if (chan == 1) {
    /* Mono */
    u_int16 left = samples;
    while (left) {
      int toRead = MIN(left, STB/2);
      fread(stb, sizeof(s_int16), toRead*2, stdaudioin);
      if (channelMode == chmRight) {
	ringcpyXX((u_int16 *)stb  , 2, (u_int16 *)stb+1, 2, toRead);
      } else if (channelMode == chmMono) {
	int i;
	s_int16 *tp = stb;
	for (i=0; i<toRead; i++) {
	  s_int16 t = Sat32To16(((s_int32)tp[0]+(s_int32)tp[1])>>1);
	  *tp++ = t; *tp++ = t;
	}
      }
      /* Now that left channel contains correct data, copy that to result */
      ringcpyXX((u_int16 *)data[0], 1, (u_int16 *)stb, 2, toRead);
      /* Process data */
#ifdef AMPLIFY_SIGNAL_BY
      Gain(data[0], toRead,
	(u_int16)(AMPLIFY_SIGNAL_BY >= 64.0 ? 65535U : AMPLIFY_SIGNAL_BY*1024));
#endif /* AMPLIFY_SIGNAL_BY */
#if defined(USE_SUBSONIC) || defined(USE_AGC)
      LibSubsonic(agcLib, subsonicM, &subsonicG, data[0], toRead);
#endif /* USE_SUBSONIC || USE_AGC */
#ifdef USE_AGC
      LibAgc(agcLib, agc, &agcConsts, data[0], toRead);
#endif /* USE_AGC */

      /* Copy back to stereo ring buffer so we can play it. */
      ringcpyXX((u_int16 *)stb  , 2, (u_int16 *)data[0], 1, toRead);
      ringcpyXX((u_int16 *)stb+1, 2, (u_int16 *)data[0], 1, toRead);
      fwrite(stb, sizeof(s_int16), toRead*2, stdaudioout);
      if (clearVuMeter) {
	clearVuMeter = 0;
	vuMeterMemory[0] = vuMeterMemory[1] = 0;
      }
      VuMeter(data[0], toRead, vuMeterMemory);
      if (es->cancel || !pauseMode) {
	data[0] += toRead;
	left -= toRead;
      }
    }
  } else {
    u_int16 left = samples;
    while (left) {
      int toRead = MIN(left, STB/2);
      fread(stb, sizeof(s_int16), toRead*2, stdaudioin);
#ifdef AMPLIFY_SIGNAL_BY
      Gain(stb, toRead*2,
	(u_int16)(AMPLIFY_SIGNAL_BY >= 64.0 ? 65535U : AMPLIFY_SIGNAL_BY*1024));
#endif
      ringcpyXX((u_int16 *)data[0], 1, (u_int16 *)stb  , 2, toRead);
      ringcpyXX((u_int16 *)data[1], 1, (u_int16 *)stb+1, 2, toRead);
#if defined(USE_SUBSONIC) || defined(USE_AGC)
      LibSubsonic(agcLib, subsonicM  , &subsonicG, data[0], toRead);
      LibSubsonic(agcLib, subsonicM+1, &subsonicG, data[1], toRead);
#endif /* USE_SUBSONIC || USE_AGC */
#ifdef USE_AGC
      LibAgc(agcLib, agc  , &agcConsts, data[0], toRead);
      LibAgc(agcLib, agc+1, &agcConsts, data[1], toRead);
#endif /* USE_AGC */
#if defined(USE_SUBSONIC) || defined(USE_AGC)
      ringcpyXX((u_int16 *)stb  , 2, (u_int16 *)data[0], 1, toRead);
      ringcpyXX((u_int16 *)stb+1, 2, (u_int16 *)data[1], 1, toRead);
#endif /* USE_SUBSONIC || USE_AGC */
      fwrite(stb, sizeof(s_int16), toRead*2, stdaudioout);
      if (clearVuMeter) {
	clearVuMeter = 0;
        vuMeterMemory[0] = vuMeterMemory[1] = 0;
      }
      VuMeter(data[0], toRead, vuMeterMemory  );
      VuMeter(data[1], toRead, vuMeterMemory+1);
      if (es->cancel || !pauseMode) {
	data[0] += toRead;
	data[1] += toRead;
	left -= toRead;
      }
    }
  }

  return samples;
}




/**
   struct EncoderServices service ESOutput(). Writes data to MP3 file.

   \param es Encoder Service structure.
   \param data Interleaved audio data. This data buffer may
        be used as working space by Output().
   \param bytes How many bytes of data to write.
*/
s_int16 ESOutput(struct EncoderServices *es, u_int16 *data, s_int16 bytes) {
  if (oFp->op->Write(oFp, data, 0, bytes) == bytes) {
    return 0;
  }
  return -1;
}



const u_int16 linToDBTab[5] = {36781, 41285, 46341, 52016, 58386};

/*
  Converts a linear 16-bit value between 0..65535 to decibels.
    Reference level: 32768 = 96dB (32767 = 95dB).
  Bugs:
    - For the input of 0, 0 dB is returned, because minus infinity cannot
      be represented with integers.
    - Assumes a ratio of 2 is 6 dB, when it actually is approx. 6.02 dB.
*/
auto u_int16 LinToDB(u_int16 n) {
  int res = 96, i;

  if (!n)               /* No signal should return minus infinity */
    return 0;

  while (n < 32768U) {  /* Amplify weak signals */
    res -= 6;
    n <<= 1;
  }

  for (i=0; i<5; i++)   /* Find exact scale */
    if (n >= linToDBTab[i])
      res++;

  return res;
}


#if 0
const u_int16 DBToLinTab[6] = {
#if 0
  /* Correct rounding, so perfect complement to LinToDB() */
  34716, 38968, 43740, 49097, 55109, 61858
#else
  /* Gives nicer powers of two but does not exactly complement LinToDB() */
  32768, 36781, 41285, 46341, 52016, 58386
#endif
};

auto u_int16 DBToLin(u_int16 n) {
  u_int16 res = 1;
  if (n < 6) {
    return 0;
  }
  while ((n-=6) >= 6) {
    res <<= 1;
  }
  res = ((u_int32)res * DBToLinTab[n] + (1<<14)) >> 15;
  return res;
}
#endif





enum EncoderError encErrCode = eeOk;
const char *encErrStr = "";

/*
  This is the encoder task. Because it "only" calls the MP3 Encoder
  function, it seems very trivial. However, this tasks does all the
  encoding, and the encoder also calls ESRead() and ESOutput()
  functions through the EncoderServices structure in enc.
*/
void EncoderTask(void) {
  encErrCode = enc->Encode((struct Encoder *)enc, &encErrStr);
}


int channels = 1;

struct EncoderServices es;

#define MAX_CMD_LEN 80
char cmd[MAX_CMD_LEN] = "";


ioresult main(char *parameters) {
  int nParam, i;
  char *p = parameters;
  int retVal = EXIT_SUCCESS;
  u_int32 lastSeconds = 0, sampleRate = 0;
  void *fileBufLib = NULL, *encLib = NULL;
  struct AutoBufferElement *abe = NULL;
  int fileNumber = 0;
  char *fileName = NULL;
  int nextFile = 1;
  int oldPauseMode;
  static s_int16 cmdLen = -1;
  int useMp3 = 1;
  s_int16 bitRate = 0;
  s_int16 quality = 7;

  nParam = RunLibraryFunction("ParamSpl", ENTRY_MAIN, (u_int16)parameters);
  for (i=0; i<nParam; i++) {
    if (!strcmp(p, "-h")) {
      printf("Usage: Rec [-fm|-fo|-cs|-cm|-cl|-cr|-q{x}|-b{x}|-r{x}|-h] outFile\n"
	     "-fm|-fo\tFormat MP3|Ogg Vorbis\n"
	     "-cs|-cm|-cl|-cr\tStereo|Mono|Left|Right\n"
	     "-qx\tSet quality to x (0-10, higher is better)\n"
	     "-bx\tSet bitrate to x kbit/s (1-511)\n"
	     "-rx\tSet sample rate to x Hz, or x kHz if x<1000\n"
	     "-h\tShow this help\n"
	     "outFile\tThe output file name (e.g. D:REC.MP3)\n");
      goto finally;
    } else if (!strcmp(p, "-fm")) {
      useMp3 = 1;
    } else if (!strcmp(p, "-fo")) {
      useMp3 = 0;
    } else if (!strcmp(p, "-cs")) {
      channelMode = chmJointStereo;
    } else if (!strcmp(p, "-cm")) {
      channelMode = chmMono;
    } else if (!strcmp(p, "-cl")) {
      channelMode = chmLeft;
    } else if (!strcmp(p, "-cr")) {
      channelMode = chmRight;
    } else if (!strncmp(p, "-q", 2)) {
      int t = atoi(p+2);
      if (t >= 0 && t <= 10) {
	quality = t;
	bitRate = 0;
      }
    } else if (!strncmp(p, "-b", 2)) {
      int t = atoi(p+2);
      if (t >= 0 && t <= 511) {
	bitRate = t;
	quality = -1;
      }
    } else if (!strncmp(p, "-r", 2)) {
      s_int32 t = strtol(p+2, NULL, 0);
      if (t > 0) {
	if (t < 1000) {
	  t *= 1000;
	}
	sampleRate = t;
      }
    } else if (!fileName) {
      fileName = p;
    } else {
      printf("Unknown parameter \"%s\", ignoring\n", p);
    }
    p += strlen(p)+1;
  }

  channels = (channelMode == chmJointStereo) ? 2 : 1;

  if (!fileName) {
    printf("!noFileName\n");
    retVal = EXIT_FAILURE;
    goto finally;
  }

  fileBufLib = LoadLibrary("filebuf");
  if (!fileBufLib) {
    printf("!FILEBUF.DL3\n");
    retVal = EXIT_FAILURE;
    goto finally;
  }

  if (useMp3) {
    encLib = LoadLibrary("encmp3");
    if (!encLib) {
      printf("!ENCMP3.DL3\n");
      retVal = EXIT_FAILURE;
      goto finally;
    }
  } else {
    encLib = LoadLibrary("encvorb");
    if (!encLib) {
      printf("!ENCVORB.DL3\n");
      retVal = EXIT_FAILURE;
      goto finally;
    }
  }

  if (quality < 0 && !bitRate) {
    quality = 7;
  }

#ifdef USE_MIC_INPUTS
  ioctl(stdaudioin, IOCTL_AUDIO_SELECT_INPUT, (void *)(AID_MIC1|AID_MIC2));
#endif

  /* Make file buffer at least 1024 stereo samples long to make sure
     sound is uninterrupted even while the encoder is running. */
  if (ioctl(stdaudioin, IOCTL_AUDIO_GET_INPUT_BUFFER_SIZE, NULL) < 2048) {
    ioctl(stdaudioin, IOCTL_AUDIO_SET_INPUT_BUFFER_SIZE, (void *)2048);
  }


  /* Note that the input sample rate can be adjusted only with a few
     steps, so you after setting the sample rate you need to reread
     what you actually got. */
  if (sampleRate) {
    ioctl(stdaudioin, IOCTL_AUDIO_SET_IRATE, (void *)(&sampleRate));
  }
  ioctl(stdaudioin, IOCTL_AUDIO_GET_IRATE, (void *)(&sampleRate));

  if (bitRate) {
    printf("Recording %ld Hz %s at %d kbit/s\n",
      sampleRate, (channels == 1) ? "mono" : "stereo",
      bitRate);
  } else {
    printf("Recording %ld Hz %s at quality %d\n",
	   sampleRate, (channels == 1) ? "mono" : "stereo",
	   quality);
  }

  ioctl(stdaudioout, IOCTL_AUDIO_SET_ORATE, &sampleRate);


#if defined(USE_SUBSONIC) || defined(USE_AGC)
  agcLib = LoadLibrary("agc");
  if (!agcLib) {
    printf("!AGC.DL3\n");
    retVal = EXIT_FAILURE;
    goto finally;
  }

  LibInitSubsonic(agcLib, subsonicM, channels, &subsonicG, sampleRate);
#ifdef USE_AGC
  {
    u_int16 maxGain = (AGC_MAX_GAIN >= 64.0) ? 65535U : AGC_MAX_GAIN*1024;
    LibInitAgc(agcLib, agc, channels, &agcConsts, maxGain, sampleRate);
  }
#endif /* USE_AGC */
#endif /* USE_SUBSONIC || USE_AGC */

  while (nextFile) {
    printf("~%04x'%s\n", UIMSG_TEXT_SHORT_FILE_NAME, fileName);

    oFp = fopen(fileName, "wb");
    if (!oFp) {
      printf("!cantOpenWriteFile:%s\n", fileName);
      retVal = EXIT_FAILURE;
      goto finally;
    }

    /* Prepare EncoderServices structure */
    memset(&es, 0, sizeof(es));
    es.Read = ESRead;
    es.Output = ESOutput;
    pauseMode = 0;
    oldPauseMode = 2;

    /* Create the encoder (same call convention for both MP3 and Ogg Vorbis
       formats, so no need to have separate code for different encoders. */

    enc = LibEncMp3Create(encLib, &es, channels, sampleRate, 
			  bitRate ?
			  (ENCODER_QUALITY_CBR|ENCODER_QUALITY_MULT1K|bitRate):
			  (ENCODER_QUALITY_QUALITY|quality));

    if (!enc) {
      printf("!createEnc\n");
      retVal = EXIT_FAILURE;
      goto finally;
    }

    /* Attach file buffer. Make it as big as possible upto 49152 bytes. */
    {
      u_int16 bufferSize = 24576; /* In 16-bit words */
      while (bufferSize >= 2048 &&
	     !(abe = CreateFileBuf(fileBufLib, oFp, bufferSize))) {
	bufferSize -= 1024;
      }
      if (!abe) {
	printf("!outputToFileBuf!\n");
	retVal = EXIT_FAILURE;
	goto finally;
      }
      printf("Output buffer size %u bytes\n", bufferSize*2);
    }

#if 0
    /* Demo for setting limited frame length for Ogg Vorbis.
       This code contains a bit of "poke" code magic and is not
       as such a very nice example. But until we have official
       support for setting the Ogg frame maximum # of samples, it
       will do. */
#ifdef USE_OGG
    {
      u_int16 *maxSamplesInFrame = ((u_int16 *)enc) + 39;
      /* When encoder has more than maxSamplesInFrame samples in one
	 Ogg frame, it will finalize it and create a new one. The
	 number may be between 128-1 and 61440. Default is 61440. */
      *maxSamplesInFrame = 2048-1;
    }
#endif /* USE_OGG */
#endif
    /* Start the encoder task function EncoderTask(). TASK_DECODER
       starts as a default at a higher priority than this main task. */
    StartTask(TASK_DECODER, EncoderTask);

    /* Main loop runs as long as EncoderTask() has not ended.
       The main loop will take care of writing the MP3/Vorbis data
       EncoderTask() has created into the ring buffer. Also the user
       interface is handled here. */

    while (pSysTasks[TASK_DECODER].task.tc_State &&
	   pSysTasks[TASK_DECODER].task.tc_State != TS_REMOVED) {
      static u_int32 totalIoBlocked = 0;
      int hadAnythingToDo = 0;

      if (appFlags & APP_FLAG_QUIT) {
	es.cancel = 1; // Asks the encoder task to stop.
	nextFile = 0;
      }

      if (ioctl(stdin, IOCTL_TEST, NULL) > 0) {
	char c = fgetc(stdin);
	int targetOffset = 0;

	hadAnythingToDo = 1;

	if (cmdLen < 0) {
	  switch (c) {
	  case ' ':
	    pauseMode = !pauseMode;
	    break;
	  case 'q':
	    es.cancel = 1; // Asks the encoder task to stop.
	    nextFile = 0;
	    break;
	  case 'n':
	    es.cancel = 1;
#if 0
	    nextFile = 1;
#else
	    nextFile = 0;
#endif
	    break;
	  case '~':
	    cmdLen = 0;
	    cmd[0] = '\0';
	    if (appFlags & APP_FLAG_ECHO) {
	      putchar('~');
	    }
	    break;
	  } /* switch(c) */
	} else { /* cmdLen >= 0 */
	  if (c==8 || c==0x7f) {
	    if (appFlags & APP_FLAG_ECHO) {
	      printf("%c[1D",27);
	      printf("%c[0K",27);
	    }
	    if (--cmdLen >= 0) {
	      cmd[cmdLen] = '\0';
	    }
	  } else {
	    if (appFlags & APP_FLAG_ECHO) {
	      putchar(c);
	    }
	    if (c != '\n') {
	      cmd[cmdLen] = c;
	      if (cmdLen < MAX_CMD_LEN-1) {
		cmdLen++;
	      }
	      cmd[cmdLen] = '\0';
	    } else { /* c == '\n' */
	      int i;
	      u_int16 message = strtol(cmd+0, NULL, 16);
	      if (cmdLen == 4) {
		if (message == UIMSG_U32_FILE_POSITION) {
		  printf("~%04x=%lu\n", UIMSG_U32_FILE_POSITION, ftell(oFp));
		} else if (message == UIMSG_U32_BITS_PER_SECOND) {
		  printf("~%04x=%lu\n", UIMSG_U32_BITS_PER_SECOND,
			 es.avgBitRate);
		} else if (message == UIMSG_S16_MAX_VU) {
		  u_int16 vuLin;
		  s_int16 vuDB;
		  vuLin = MAX(vuMeterMemory[0], vuMeterMemory[1]);
		  clearVuMeter = 1;
		  vuDB = LinToDB(vuLin)-95;
		  if (vuDB > 0) {
		    vuDB = 0;
		  }
		  printf("~%04x=%d\n", UIMSG_S16_MAX_VU, vuDB);
		} else {
		  printf("!cmdUnknown\n");
		}
	      } else if (cmdLen < 6 || (cmd[4] != '=' && cmd[4] != '\'')) {
		printf("!cmdLine\n");
	      } else {
		s_int32 val = strtol(cmd+5, NULL, 0);
		if (message == UIMSG_BUT_PAUSE) {
		  pauseMode = val;
		} else {
		  printf("!cmdUnknown\n");
		}
	      } /* c == '\n' */
	      cmdLen = -1;
	    }
	  }
	}
      } /* if (ioctl(stdin, IOCTL_TEST, NULL) > 0) */
      
      /* Check that the output file has not grown close to VSOS's
	 capabilities to handle -> limit file size to just over
	 1 billion bytes, then autojump to a new file number. */
      if (ftell(oFp) > 1000000000UL) {
	hadAnythingToDo = 1;
	es.cancel = 1;
#if 0
	nextFile = 1;
#else
	nextFile = 0;
#endif
      }

      if (pauseMode != oldPauseMode) {
	hadAnythingToDo = 1;
	if (oldPauseMode < 2) {
	  printf("~%04x=%d\n", UIMSG_BUT_PAUSE, pauseMode);
	}
	oldPauseMode = pauseMode;
      }

      if (abe->ioBlocked != totalIoBlocked) {
	static char s[40];
	hadAnythingToDo = 1;
	totalIoBlocked = abe->ioBlocked;
	printf(s, "WARNING: %5.3f SECS AUDIO MISSING\n",
	       abe->ioBlocked*0.001);
      }

      if (es.playTimeSeconds != lastSeconds &&
	  (cmdLen < 0 || !(appFlags & APP_FLAG_ECHO))) {
	double br = es.avgBitRate*0.001;

	hadAnythingToDo = 1;
	lastSeconds = es.playTimeSeconds;
	  
	printf("~%04x=%lu\n", UIMSG_U32_PLAY_TIME_SECONDS, lastSeconds);
      }

      if (!hadAnythingToDo) {
	Delay(20);
      }
    } /* while (encoder task running) */

    if (enc) {
      if (encErrCode > 0) {
	printf("!errCode:%d,errStr:\"%s\"\n", encErrCode, encErrStr);
	retVal = EXIT_FAILURE;
      }

      //    printf("Delete encoder\n");
      enc->Delete((struct Encoder *)enc);
      enc = NULL;
    }

    if (oFp) {
      fclose(oFp);
      oFp = NULL;
    }
  } /* while (nextFile) */

 finally:
#if defined(USE_SUBSONIC) || defined(USE_AGC)
  if (agcLib) {
    DropLibrary(agcLib);
    agcLib = NULL;
  }
#endif /* USE_SUBSONIC || USE_AGC */

  if (encLib) {
    if (enc) {
      enc->Delete((struct Encoder *)enc);
      enc = NULL;
    }
    DropLibrary(encLib);
    encLib = NULL;
  }

  if (oFp) {
    fclose(oFp);
    oFp = NULL;
  }

  if (fileBufLib) {
    DropLibrary(fileBufLib);
    fileBufLib = NULL;
  }

  printf("~%04x=%d\n", UIMSG_BUT_DECODER_CLOSED, 1);

  return retVal;
}
