/*

  SetClock sets the clock frequency of VS1005g.
  
*/
#include <vo_stdio.h>
#include <stdlib.h>
#include <apploader.h>
#include <consolestate.h>
#include <string.h>
#include <stdlib.h>
#include <audio.h>
#include <clockspeed.h>

auto s_int16 VSOS_SetClockSpeed(register u_int32 clkHz);
auto void FindVCOLock(void);

auto s_int16 MySetClockSpeed(register u_int32 clkHz) {
  int shiftRight = clockSpeed.xtalPrescaler /*+ clockSpeed.pllPreDiv*/ +
    clockSpeed.div2 + clockSpeed.div256*8 + 1;
  int newMultiplier = 2;
  u_int32 f;

#if 0
  printf("SetClockSpeed(%ld)\n", clkHz);
  printf(" xtalPre %d, pllPreDiv %d, div2 %d, div256 %d\n",
	 clockSpeed.xtalPrescaler, clockSpeed.pllPreDiv,
	 clockSpeed.div2, clockSpeed.div256);
#endif

  if (clockSpeed.masterClkSrc == MASTER_CLK_SRC_RF) {
    PERIP(CLK_CF) &= ~CLK_CF_FORCEPLL;     // clock switch pois PLL-asennosta
    PERIP(CLK_CF) &= ~CLK_CF_USBCLK;    // USB clk pois plt
    clockSpeed.masterClkSrc = MASTER_CLK_SRC_XTALI;
  }

  if (clkHz & SET_CLOCK_USB) {
    if (clockSpeed.extClkKHz == 12288) {
      PERIP(CLK_CF) = 0;               // 1x kellolla

#if 0
      PERIP(ANA_CF3) = 0x8008;         //  ena_480, vcodiv=fmdiv=0,
                                      //  cntr guess 8
      PERIP(ANA_CF2) = 0x0500 | 0x18;  //  enableusb, vcopd, Vref
      PERIP(ANA_CF0)  = 0;              //  no usb stuff
      PERIP(ANA_CF1)  = 0;              // no usb stuff
#else
      PERIP(ANA_CF3) &= 0x83f8;         //  ena_480, vcodiv=fmdiv=0,
      PERIP(ANA_CF3) |= 0x8008;        //  cntr guess 8
      PERIP(ANA_CF2) &= ~0x0400;
      PERIP(ANA_CF2) |= 0x0500 | 0x18;  //  enableusb, vcopd, Vref
      //      PERIP(ANA_CF0)  = 0;              //  no usb stuff
      //      PERIP(ANA_CF1)  = 0;              // no usb stuff
#endif

      PERIP(FMCCF_LO)  = 0xffff;      // => 60.000MHz USB @ 12.288MHz
      PERIP(FMCCF_HI)  = 0x0387;      //
      PERIP(FM_CF)      = 0x0041;      // clearregs off, FM ena on

      FindVCOLock();                     // hakee oikean vco cntr arvon:
                                      //  0xF, 0xE, 0xD, ...
      // Kello corelle:
      PERIP(CLK_CF) =  0x0001;    // 2x clk jtetn pll:n
      PERIP(CLK_CF) |= (1<<6);    // select usb clk (knt muxin core
                                    // PLL:lt usb kellolle)
      PERIP(CLK_CF) |= (1<<5);    // clock switch pll asentoon

      clockSpeed.masterClkSrc = MASTER_CLK_SRC_RF;
      clockSpeed.cpuClkHz = 60e6;

      //      printf("SetClockSpeed(USB, 12288)\n");

      return 0;
    } else if (clockSpeed.extClkKHz == 12000) {

      //	printf("SetClockSpeed(USB, 12000)\n");

      clkHz = 60e6;
    } else {
      //      printf("SetClockSpeed(USB, **FAIL**)\n");

      return -1;
    }
  }


  while ((f=((s_int32)clockSpeed.extClkKHz*1000>>shiftRight)*newMultiplier) <
	 clkHz && f < clockSpeed.speedLimitHz && newMultiplier < 16) {
#if 0
    printf("  clkHz %ld, f = %ld\n", clkHz, f);
#endif
    newMultiplier++;
  }
  if (f > clockSpeed.speedLimitHz)
    newMultiplier--;

#if 0
  printf("SetClockSpeed: newMultiplier %d, shR %d\n",
      newMultiplier, shiftRight);
#endif

  ForceClockMultiplier(newMultiplier);
}




ioresult main(char *parameters) {
  int nParam, i;
  char *p = parameters;
  ioresult ret = S_OK;
  long newMHz = 0, speedLimitMHz = 0;
  int nothingToDo = 1, nextIsLimit = 0;
  int verbose = 0;

  nParam = RunLibraryFunction("ParamSpl", ENTRY_MAIN, (u_int16)parameters);
  for (i=0; i<nParam; i++) {
    if (!strcmp(p, "-h")) {
      printf("Usage: SetClock -v [f|-l f|-h]\n"
	     "-v\tBe verbose\n"
	     "f\tFrequency in MHz\n"
	     "-l f\tLimit top speed in MHz (USE WITH CAUTION!)\n"
	     "usb\tSet exactly 60 MHz\n"
	     "-h\tShow this help\n");
      goto finally;
    } else if (!strcmp(p, "-v")) {
      verbose = 1;
    } else if (!strcmp(p, "-l")) {
      nextIsLimit = 1;
    } else if (!strcmp(p, "usb")) {
      newMHz = -1;
    } else if (strtol(p, NULL, 0) > 0) {
      s_int32 t = strtol(p, NULL, 0);
      if (nextIsLimit) {
	nextIsLimit = 0;
	speedLimitMHz = t;
      } else {
	newMHz = t;
      }
    } else {
      printf("E: Unknown parameter \"%s\"\n", p);
      ret = S_ERROR;
      goto finally;
    }
    p += strlen(p)+1;
  }

  if (speedLimitMHz) {
    clockSpeed.speedLimitHz = 1000000*speedLimitMHz;
  }

  if (newMHz) {
    MySetClockSpeed(newMHz < 0 ? SET_CLOCK_USB : 1000000*newMHz);
    nothingToDo = 0;
  }

  if (nothingToDo || verbose) {
    printf("SetClock: Frequency %5.2f MHz, max %5.2f MHz\n",
	   clockSpeed.cpuClkHz*0.000001,
	   clockSpeed.speedLimitHz*0.000001);
  }

#if 0
  printf("cpuClkHz     %8ld Hz\n", clockSpeed.cpuClkHz);
  printf("peripClkHz   %8ld Hz\n", clockSpeed.peripClkHz);
  printf("speedLimitHz %8ld Hz\n", clockSpeed.speedLimitHz);
  printf("extClkKHz    %d kHz\n", clockSpeed.extClkKHz);
  printf("time         %ld ticks\n", ReadTimeCount());
  {
    u_int32 startTime = ReadTimeCount();
    u_int32 loops = 0;
    while (ReadTimeCount() == startTime)
      ;
    startTime = ReadTimeCount();
    while (ReadTimeCount()-startTime < 25) {
      loops++;
    }
    printf("%ld loops in 25 ms\n", loops);
  }
#endif


 finally:
  return ret;
}
