 /* For free support for VSIDE, please visit www.vsdsp-forum.com */

/*
  Audio decoder AUDIODEC.DL3 source code v1.03 2015-03-27
*/
#include <vo_stdio.h>
#include <volink.h>     // Linker directives like DLLENTRY
#include <apploader.h>  // RunLibraryFunction etc
#include <string.h>
#include <codecmpg.h>
#include <libdecvorb.h>
#include <libdecwav.h>
#include <sysmemory.h>
#include <libaudiodec.h>
#include "codecservices.h"
#include "ungetbuffer.h"


struct Codec *cod = NULL;
void (*Delete)(struct Codec *cod) = NULL;

void fini(void);

#define MAKE_ID(a,b,c,d) (((u_int32)(a)<<24)|((u_int32)(b)<<16)|((c)<<8)|(d))
#define MAKE_ID_MIX(a,b,c,d) (((u_int32)(c)<<24)|((u_int32)(d)<<16)|((a)<<8)|(b))

const char * __mem_y libNames[] = {NULL, "decmp3", "decwav", "decvorb",
		"decflac", "decalac", "decwma", "decaac",
		"decmidi"};

int MyDeleteAudioDecoder(register AUDIO_DECODER *dec);

DLLENTRY(MyCreateAudioDecoder)
AUDIO_DECODER *MyCreateAudioDecoder(register FILE *inFp,
				       register FILE *outFp,
				       register void (*callback)
				       (AUDIO_DECODER *dec, u_int16 samples),
				       register enum AuDecFormat auDecFormat) {
	AUDIO_DECODER *dec = calloc(1, sizeof(AUDIO_DECODER));
	const char *decoderLibName = NULL;
	if (!dec) {
		return NULL;
	}

#if 0
	fprintf(stderr, "CreateAudioDecoder\n", dec);
#endif

	{
		u_int32 fourBytes[2];
	restart:
		fread(ungetBuffer, sizeof(ungetBuffer[0])*4, 1, inFp);
		memcpy(fourBytes, ungetBuffer, sizeof(ungetBuffer));
		ungetBufferFilePos = inFp->pos;

		while (auDecFormat == auDecFGuess) {
			if ((fourBytes[0] & 0xFF00FFFFUL) == MAKE_ID_MIX('I','D','3',0)) {
				/* ID3v2 tag; let's skip it. */
				u_int32 id3Size;
				id3Size  = ((u_int32)ungetBuffer[3]&0xFF00U) << 21;
				id3Size |= ((u_int32)ungetBuffer[3]&0x00FFU) << 14;
				id3Size |= fgetc(inFp) << 7;
				id3Size |= fgetc(inFp);
				fseek(inFp, id3Size, SEEK_CUR);
				goto restart;
			} else if (fourBytes[0] == MAKE_ID_MIX('R','I','F','F')) {
				auDecFormat = auDecFWav;
			} else if (fourBytes[0] == MAKE_ID_MIX('O','g','g','S')) {
				auDecFormat = auDecFOgg;
			} else if (fourBytes[0] == MAKE_ID_MIX('f','L','a','C')) {
				auDecFormat = auDecFFlac;
			} else if (fourBytes[0] == MAKE_ID_MIX('c','a','f','f')) {
				auDecFormat = auDecFAlac;
			} else if (fourBytes[0] == MAKE_ID_MIX('0','&',178,'u')) {
				auDecFormat = auDecFWma;
			} else if (fourBytes[0] == MAKE_ID_MIX('A','D','I','F') ||
						fourBytes[0] == MAKE_ID_MIX('f','t','y','p') ||
						fourBytes[0] == MAKE_ID_MIX('m','o','o','v') ||
						(((u_int16)fourBytes[0] & 0xFFF6U) == 0xFFF0U)) {
				auDecFormat = auDecFAac;
			} else if (fourBytes[0] == MAKE_ID_MIX('M','T','h','d')) {
				auDecFormat = auDecFMidi;
			} else if (((u_int16)fourBytes[0] & 0xFFE6U) == 0xFFE2U) {
				auDecFormat = auDecFMp3;
			} else {
				fourBytes[0] = fourBytes[1];
				fourBytes[1] = 0x67fba112;
				if (fourBytes[0] == 0x67fba112) {
					auDecFormat = auDecFMp3;
				}
			}
		}
	}

	dec->format = auDecFormat;

	if (auDecFormat < sizeof(libNames)/sizeof(libNames[0])) {
		decoderLibName = libNames[auDecFormat];
	} else {
		MyDeleteAudioDecoder(dec);
		return NULL;
	}

	if (decoderLibName) {
		// Make generic decoder load operation
		//fprintf(stderr, "# Try load %s\n", decoderLibName);
		dec->decoderLib = LoadLibrary(decoderLibName);
		//fprintf(stderr, "# Lib = %p\n", dec->decoderLib);
		if (!dec->decoderLib) {
			MyDeleteAudioDecoder(dec);
			return NULL;
		}
		dec->cod = LibCodGenericCreate(dec->decoderLib);
		if (dec->cod && auDecFormat == auDecFMp3) {
			CodMpgSetTags(dec->cod, codMpgDisableCrc, codMpgEnableLayer2, codMpgLeaveOnError, 0);
		}
	}

	if (!dec->cod) {
		MyDeleteAudioDecoder(dec);
		return NULL;
	}
	memcpyYX(&dec->cs, &audioDecoderDefaultCodecServices,
			sizeof(audioDecoderDefaultCodecServices));

	dec->cod->cs = &dec->cs;
	dec->inFp = inFp;
	dec->outFp = outFp;
	{
		u_int32 oldPos = dec->inFp->pos - 8;
		if (!fseek(dec->inFp, 0, SEEK_END)) {
			dec->cs.fileSize = dec->cs.fileLeft = ftell(inFp);
			fseek(inFp, oldPos, SEEK_SET);
		} else {
			dec->cs.fileSize = dec->cs.fileLeft = 0xFFFFFFFFU;
			ungetBufferPointer = 0;
		}
	}
	dec->callback = callback;

	return dec;
}

DLLENTRY(MyDeleteAudioDecoder)
int MyDeleteAudioDecoder(register AUDIO_DECODER *dec) {
	if (dec) {
		if (dec->cod && dec->cod->Delete) {
			dec->cod->Delete(dec->cod);
		}
		if (dec->decoderLib) {
			DropLibrary(dec->decoderLib);
		}
		free(dec);
	}

	return 0;
}


DLLENTRY(MyDecodeAudio)
u_int16 MyDecodeAudio(register AUDIO_DECODER *dec,
			 register const char **eStr) {
	int errCode;
	//	printf("#soon decode\n");
	errCode = dec->cod->Decode(dec->cod, &dec->cs, eStr);
	//	printf("#decode returned %d, \"%s\"\n", errCode, *eStr);
	return errCode;
}
